import numpy as np
import pyomo.environ as pyo

from pyomo.core.expr import numeric_expr
from pyomo.opt import SolverFactory
from sklearn.preprocessing import StandardScaler
from penlm.base_estimators import BaseClassifier, BaseRegressor
from abc import ABC, abstractmethod
from typing import Dict, Tuple


class BaseSmoothLinear(ABC):
    penalty_types = ['d1','d2']  

    
    def set_parameters(self,
                       parameters: Dict):
        self.lambd = parameters['lambda']
        self.parameters['lambda'] = self.lambd
        if self.lambd <= 0:
            raise ValueError('Lambda <= 0')


    def _solve_optim(self,
                     X: np.ndarray,
                     Y: np.ndarray) -> Tuple[int, np.ndarray]:
        P = X.shape[1]
        N = X.shape[0]
        model = pyo.ConcreteModel()
        model.P = pyo.RangeSet(0, P-1)
        model.beta = pyo.Var(model.P, 
                             domain = pyo.Reals)
        if self.fit_intercept:
            model.intercept = pyo.Var(domain = pyo.Reals,
                                      initialize = 0)
        for p in range(P):
            model.beta[p]= 0
        
        model.obj = pyo.Objective(expr = self._obj_expr(model,X,Y), 
                                  sense = pyo.minimize)
        opt = SolverFactory(self.solver,
                            options = {'max_iter': self.max_iter}) 
        opt.solve(model)
        
        beta = []
        for p in range(P):
            beta.append(pyo.value(model.beta[p]))   
        beta = np.array(beta)
        if self.fit_intercept:
            intercept = pyo.value(model.intercept)
        else:
            intercept = 0
        return intercept, beta


    @abstractmethod
    def _obj_expr(self,
                  model: pyo.ConcreteModel,
                  X: np.ndarray,
                  Y: np.ndarray) -> numeric_expr:
        pass
       
        
    def _get_penalty_expr(self,
                          model: pyo.ConcreteModel,
                          P: int) -> numeric_expr:
        pen = None
        if self.penalty_type=='d2':
            pen = P*P*sum((model.beta[p] - 2*model.beta[p+1] + model.beta[p+2])**2 for p in range(P-2))
        elif self.penalty_type=='d1':
            pen = P*sum((model.beta[p] - model.beta[p+1])**2 for p in range(P-1))                
        return pen 
        
        
                                                                                  
class SmoothLinearClassifier(BaseSmoothLinear, BaseClassifier):
    def __init__(self,
                 solver: str = 'ipopt',
                 penalty_type: str = 'd2',
                 fit_intercept: bool = True,
                 scale: bool = True,
                 max_iter: int = 100,
                 class_weight: str = 'balanced'):
        if penalty_type in self.penalty_types:
            self.penalty_type = penalty_type
        else:
            raise ValueError(f'Illegal Penalization Type == {penalty_type}')     
        self.solver = solver
        self.fit_intercept = fit_intercept
        self.class_weight = class_weight
        self.parameters = {}
        self.lambd = None
        self.max_iter = max_iter
        self.beta = []
        self.intercept = []
        self.scale = scale
        self.scaler = None
        self.classes = None


    def fit(self,
            X: np.ndarray,
            Y: np.ndarray) -> 'SmoothLinearClassifier': 
        if self.scale:
            self.scaler = StandardScaler()
            X = self.scaler.fit_transform(X)            
        self.classes = np.unique(Y)
        n_classes = len(self.classes)
        if n_classes < 2:
            raise ValueError('''This estimator needs samples of at least 2 classes in the data, 
                                but the data contains only one class''')
        elif n_classes == 2:
            intercept, beta = self._solve_optim(X, Y)
            self.intercept.append(intercept)
            self.beta.append(beta)
        else:
            for j,k in enumerate(self.classes):
                _Y = np.where(Y == k, 1, 0)
                intercept, beta = self._solve_optim(X, _Y)
                self.intercept.append(intercept)
                self.beta.append(beta)
        self.intercept = np.array(self.intercept)
        self.beta = np.array(self.beta)    
        return self
        
        
    def _obj_expr(self,
                  model: pyo.ConcreteModel,
                  X: np.ndarray,
                  Y: np.ndarray) -> numeric_expr:
        N = X.shape[0]
        P = X.shape[1]
        if self.class_weight=='balanced': 
            unique_weights = N/(len(np.unique(Y))*np.bincount(Y))
            weights = np.zeros(N)
            for i in np.unique(Y):
                weights[np.where(Y==i)] = unique_weights[i]
        else:
            weights = np.ones(N)
        if self.fit_intercept:
            obj = -sum(weights[j]*(Y[j]*(sum(X[j,p]*model.beta[p] for p in range(P))+model.intercept) - pyo.log(1+pyo.exp(model.intercept+sum(X[j,p]*model.beta[p] for p in range(P)))))  for j in range(N))
        else:
            obj = -sum(weights[j]*(Y[j]*(sum(X[j,p]*model.beta[p] for p in range(P))) - pyo.log(1+pyo.exp(sum(X[j,p]*model.beta[p] for p in range(P)))))  for j in range(N))
        pen = self._get_penalty_expr(model, P)
        return obj + self.lambd*pen      
                
 
 
class SmoothLinearRegressor(BaseSmoothLinear, BaseRegressor):
    def __init__(self,
                 solver: str = 'ipopt',
                 penalty_type: str = 'd2',
                 fit_intercept: bool = True,
                 scale: bool = True,
                 max_iter: int = 100):
        if penalty_type in self.penalty_types:
            self.penalty_type = penalty_type
        else:
            raise ValueError(f'Illegal Penalization Type == {penalty_type}')        
        self.solver = solver
        self.fit_intercept = fit_intercept
        self.parameters = {}
        self.lambd = None
        self.max_iter = max_iter
        self.beta = None
        self.intercept = None
        self.scale = scale
        self.scaler = None


    def fit(self,
            X: np.ndarray,
            Y: np.ndarray) -> 'SmoothLinearRegressor':    
        if self.scale:
            self.scaler = StandardScaler()
            X = self.scaler.fit_transform(X)
        self.intercept, self.beta = self._solve_optim(X, Y)
        return self
        
        
    def _obj_expr(self,
                  model: pyo.ConcreteModel,
                  X: np.ndarray,
                  Y: np.ndarray) -> numeric_expr:
        N = X.shape[0]
        P = X.shape[1]
        if self.fit_intercept:
            obj = sum((Y[j]-sum(X[j,p]*model.beta[p] for p in range(P))-model.intercept)**2 for j in range(N))
        else:
            obj = sum((Y[j]-sum(X[j,p]*model.beta[p] for p in range(P)))**2 for j in range(N))
        pen = self._get_penalty_expr(model, P)
        return obj + self.lambd*pen
        




import numpy as np
import itertools, copy, joblib

from sklearn.metrics import accuracy_score, r2_score
from typing import Dict, Callable
               
                           
class GridSearchCV:
    def __init__(self,
                 estimator: 'BaseEstimator',
                 parameters: Dict,
                 cv: 'ScikitCVObj',
                 scoring: Callable = None,
                 verbose: bool = False,
                 n_jobs: int = -1):
        self.parameters_packed = parameters
        flattened = [[(key, val) for val in values] for key, values in self.parameters_packed.items()]
        self.parameters_unpacked = [dict(items) for items in itertools.product(*flattened)]
        self.estimator = estimator
        self.cv = cv
        if scoring == None:
            if estimator.estimator_type=='classifier':
                self.scoring = accuracy_score 
            elif estimator.estimator_type=='regressor':
                self.scoring = r2_score                     
        elif callable(scoring):
            self.scoring = scoring
        else:
            raise ValueError(f'Illegal scoring {self.scoring}')
        self.verbose = verbose
        self.n_jobs = n_jobs
        self.fold_estimators = []
        self.fold_best_estimators = []
        self.fold_train_scores = []
        self.fold_test_scores = []
        self.mean_fold_train_scores = []
        self.mean_fold_test_scores = []
        self.std_fold_train_scores = []
        self.std_fold_test_scores = []
        self.best_estimator = None
        self.best_parameters = None
        self.best_estimator_train_score = None 


    def fit(self, 
            X: np.ndarray, 
            Y: np.ndarray):
        for counter,(train_index, test_index) in enumerate(self.cv.split(X,Y)):
            estimators = []
            for param_dict in self.parameters_unpacked:
                estimator = copy.deepcopy(self.estimator)
                estimator.set_parameters(param_dict) 
                estimators.append(estimator)
            estimators = joblib.Parallel(n_jobs=self.n_jobs)(joblib.delayed(estimators[i].fit)(X[train_index],Y[train_index]) for i in range(len(self.parameters_unpacked)))
            inner_train_scores = joblib.Parallel(n_jobs=self.n_jobs)(joblib.delayed(estimators[i].score)(X[train_index],Y[train_index],self.scoring) for i in range(len(self.parameters_unpacked)))
            inner_test_scores = joblib.Parallel(n_jobs=self.n_jobs)(joblib.delayed(estimators[i].score)(X[test_index],Y[test_index],self.scoring) for i in range(len(self.parameters_unpacked)))
            self.fold_estimators.append(estimators)
            self.fold_train_scores.append(inner_train_scores)
            self.fold_test_scores.append(inner_test_scores)
            best_index = inner_test_scores.index(max(inner_test_scores))
            self.fold_best_estimators.append(estimators[best_index])
            if self.verbose:
                print()
                print(' Grid Search Iternum: {}'.format(counter))
                print()
                for p,s in zip(self.parameters_unpacked,inner_test_scores):   
                    print(' Inner Parameters: {}'.format(p))
                    print(' Inner Score: {}'.format(s))
                    print()
                print()
                if len(set(inner_test_scores))==1:
                    print(' All the parameter tuples have the same score in this fold')  
        self.mean_fold_train_scores = np.mean(self.fold_train_scores, axis=0)
        self.mean_fold_test_scores = np.mean(self.fold_test_scores, axis=0)
        self.std_fold_train_scores = np.std(self.fold_train_scores, axis=0)
        self.std_fold_test_scores = np.std(self.fold_test_scores, axis=0)
        best_index = np.argmax(self.mean_fold_test_scores)
        self.best_parameters = self.parameters_unpacked[best_index]
        if self.verbose:
            print()
            print(' Best Parameters: {}'.format(self.best_parameters))
            print()     
        estimator = copy.deepcopy(self.estimator)
        estimator.set_parameters(self.best_parameters)           
        estimator.fit(X,Y)
        self.best_estimator = estimator
        self.best_estimator_train_score = estimator.score(X,Y,self.scoring)
        

    def score(self,
              X: np.ndarray,
              Y: np.ndarray) -> np.ndarray:
        return self.best_estimator.score(X, Y)
              
              
                 



import numpy as np

from sklearn.preprocessing import StandardScaler      
from sklearn.linear_model import Ridge, LogisticRegression
from sklearn.metrics import accuracy_score, r2_score
from penlm.base_estimators import BaseClassifier, BaseRegressor
from typing import Dict, Callable
                           
                                                                         
class BARClassifier(BaseClassifier):
    def __init__(self,
                 fit_intercept: bool = True,
                 max_iter: int = 5,
                 class_weight: str = 'balanced',
                 scale: bool = True,
                 random_state: int = None,
                 solver: str = 'saga',
                 scikit_max_iter: int = 1000,
                 scoring: Callable = None):
        self.class_weight = class_weight
        self.solver = solver 
        self.fit_intercept = fit_intercept
        self.parameters = {}
        self.lambd = None
        self.phi = None
        self.beta = []
        self.intercept = []
        self.scale = scale
        self.scaler = None
        self.random_state = random_state
        self.max_iter = max_iter
        self.n_iters = 0
        self.scikit_max_iter = scikit_max_iter
        if scoring == None:
            self.scoring = accuracy_score 
        elif callable(scoring):
            self.scoring = scoring 
        else:
            raise ValueError(f'Illegal scoring {self.scoring}')
        self.classes = None

    
    def set_parameters(self,parameters):
        self.lambd = parameters['lambda']
        self.parameters['lambda'] = self.lambd
        self.parameters['n_iters'] = 0
        if self.lambd <= 0:
            raise ValueError('Lambda <= 0')
                
                
    def fit(self,
            X: np.ndarray,
            Y: np.ndarray) -> 'BARClassifier':
        if self.scale:
            self.scaler = StandardScaler()
            X = self.scaler.fit_transform(X)            
        self.classes = np.unique(Y)
        n_classes = len(self.classes)
        if n_classes < 2:
            raise ValueError('''This estimator needs samples of at least 2 classes in the data, 
                                but the data contains only one class''')
        elif n_classes == 2:
            estimator = LogisticRegression(penalty = 'l2',
                                           class_weight = self.class_weight,
                                           C = 1/self.lambd,
                                           fit_intercept = self.fit_intercept,
                                           random_state = self.random_state,
                                           solver = self.solver,
                                           max_iter = self.scikit_max_iter)
            estimator.fit(X, Y)
            recipr_w = np.square(estimator.coef_[0])
            _X = np.zeros(X.shape)
            for i in range(X.shape[0]):
                _X[i,:] = X[i,:]*recipr_w
            estimator.fit(_X, Y)
            best_intercept = estimator.intercept_[0] 
            best_beta = estimator.coef_[0]*recipr_w
            score = 1 - self._pred_score(X, Y, best_beta, best_intercept)
            score_decreased = True
            k = 0
            while (k < self.max_iter) and score_decreased:
                recipr_w = np.square(best_beta)
                _X = np.zeros(X.shape)
                for i in range(X.shape[0]):
                    _X[i,:] = X[i,:]*recipr_w
                estimator.fit(_X,Y)
                beta = estimator.coef_[0]*recipr_w
                intercept = estimator.intercept_[0]
                tmp_score = 1 - self._pred_score(X, Y, beta, intercept)
                if tmp_score > score:
                    score_decreased = False
                else:
                    best_beta = beta
                    best_intercept = intercept    
                    self.n_iters += 1
                    self.parameters['n_iters'] += 1
                    k += 1
            self.intercept.append(best_intercept)
            self.beta.append(best_beta) 
        else:    
            for j,k in enumerate(self.classes):            
                estimator = LogisticRegression(penalty = 'l2',
                                               class_weight = self.class_weight,
                                               C = 1/self.lambd,
                                               fit_intercept = self.fit_intercept,
                                               random_state = self.random_state,
                                               solver = self.solver,
                                               max_iter = self.scikit_max_iter)
                _Y = np.where(Y == k, 1, 0)                                    
                estimator.fit(X, _Y)
                recipr_w = np.square(estimator.coef_[0])
                _X = np.zeros(X.shape)
                for i in range(X.shape[0]):
                    _X[i,:] = X[i,:]*recipr_w
                estimator.fit(_X, _Y)
                best_intercept = estimator.intercept_[0] 
                best_beta = estimator.coef_[0]*recipr_w
                score = 1 - self._pred_score(X, _Y, best_beta, best_intercept)
                score_decreased = True
                k = 0
                while (k < self.max_iter) and score_decreased:
                    recipr_w = np.square(best_beta)
                    _X = np.zeros(X.shape)
                    for i in range(X.shape[0]):
                        _X[i,:] = X[i,:]*recipr_w
                    estimator.fit(_X, _Y)
                    beta = estimator.coef_[0]*recipr_w
                    intercept = estimator.intercept_[0]
                    tmp_score = 1 - self._pred_score(X, _Y, beta, intercept)
                    if tmp_score > score:
                        score_decreased = False
                    else:
                        best_beta = beta
                        best_intercept = intercept    
                        self.n_iters += 1
                        self.parameters['n_iters'] += 1
                        k += 1
                self.intercept.append(best_intercept)
                self.beta.append(best_beta)
        self.intercept = np.array(self.intercept)
        self.beta = np.array(self.beta)            
        return self


    def _pred_score(self, 
                    X: np.ndarray, 
                    Y: np.ndarray, 
                    beta: np.ndarray, 
                    intercept: int) -> np.ndarray:
        pred_Y = []
        for x in X:
            if np.dot(beta,x) + intercept > 0:
                pred_Y.append(1)
            else:
                pred_Y.append(0)   
        score = self.scoring(Y, pred_Y)
        return score
        


class BARRegressor(BaseRegressor):
    def __init__(self,
                 fit_intercept: bool = True,
                 max_iter: int = 5,
                 scale: bool = True,
                 random_state: int = None,
                 scikit_max_iter: int = 1000,
                 scoring: Callable = None):
        self.fit_intercept = fit_intercept
        self.parameters = {}
        self.lambd = None
        self.phi = None
        self.beta = None
        self.intercept = None
        self.scale = scale
        self.scaler = None
        self.random_state = random_state
        self.max_iter = max_iter
        self.n_iters = 0
        self.scikit_max_iter = scikit_max_iter
        if scoring == None:
            self.scoring = r2_score
        elif callable(scoring):
            self.scoring = scoring 
        else:
            raise ValueError(f'Illegal scoring {self.scoring}')

    
    def set_parameters(self,parameters):
        self.lambd = parameters['lambda']
        self.parameters['lambda'] = self.lambd
        self.parameters['n_iters'] = 0
        if self.lambd <= 0:
            raise ValueError('Lambda <= 0')
                
                
    def fit(self,
            X: np.ndarray,
            Y: np.ndarray) -> 'BARestimator':      
        if self.scale:
            self.scaler = StandardScaler()
            X = self.scaler.fit_transform(X)
        estimator = Ridge(alpha = self.lambd,
                          fit_intercept = self.fit_intercept,
                          random_state = self.random_state,
                          max_iter = self.scikit_max_iter)
        estimator.fit(X, Y)
        recipr_w = np.square(estimator.coef_)
        _X = np.zeros(X.shape)
        for i in range(X.shape[0]):
            _X[i,:] = X[i,:]*recipr_w
        estimator.fit(_X, Y)
        self.intercept = estimator.intercept_ 
        self.beta = estimator.coef_*recipr_w
        score = self._pred_score(X, Y, self.beta, self.intercept)
        score_decreased = True
        k = 0
        while (k < self.max_iter) and score_decreased:
            recipr_w = np.square(self.beta)
            _X = np.zeros(X.shape)
            for i in range(X.shape[0]):
                _X[i,:] = X[i,:]*recipr_w
            estimator.fit(_X, Y)
            beta = estimator.coef_*recipr_w
            intercept = estimator.intercept_
            tmp_score = self._pred_score(X, Y, beta, intercept)
            if tmp_score > score:
                score_decreased = False
            else:
                self.beta = beta
                self.intercept = intercept    
                self.n_iters += 1
                self.parameters['n_iters'] += 1
                k += 1
        return self


    def _pred_score(self, 
                    X: np.ndarray, 
                    Y: np.ndarray, 
                    beta: np.ndarray, 
                    intercept: int) -> np.ndarray:
        pred_Y = []
        for x in X:
            pred = np.dot(beta,x) + intercept
            pred_Y.append(pred)
        pred_Y = np.array(pred_Y)
        score = self.scoring(Y, pred_Y)
        return score 
        
                



import numpy as np

from sklearn.preprocessing import StandardScaler      
from sklearn.linear_model import Lasso, LogisticRegression
from penlm.base_estimators import BaseClassifier, BaseRegressor
from typing import Dict
             
                                                                          
class RelaxedLassoClassifier(BaseClassifier):
    def __init__(self,
                 fit_intercept: bool = True,
                 class_weight: str = 'balanced',
                 scale: bool = True,
                 random_state: int = None,
                 solver: str = 'saga',
                 max_iter: int = 1000):
        self.class_weight = class_weight
        self.solver = solver 
        self.fit_intercept = fit_intercept
        self.parameters = {}
        self.lambd = None
        self.phi = None
        self.beta = []
        self.intercept = []
        self.scale = scale
        self.scaler = None
        self.random_state = random_state
        self.max_iter = max_iter
        self.classes = None

    
    def set_parameters(self,
                       parameters: Dict):
        self.phi = parameters['phi']
        self.lambd = parameters['lambda']
        self.parameters['phi'] = self.phi
        self.parameters['lambda'] = self.lambd
        if self.lambd <= 0:
            raise ValueError('Lambda <= 0')
        if self.phi<=0 or self.phi>1:
            raise ValueError('Phi out of (0,1]')
        
          
    def fit(self,
            X: np.ndarray,
            Y: np.ndarray) -> 'RelaxedLassoClassifier':
        if self.scale:
            self.scaler = StandardScaler()
            X = self.scaler.fit_transform(X)     
        self.classes = np.unique(Y)    
        n_classes = len(self.classes)
        if n_classes < 2:
            raise ValueError('''This estimator needs samples of at least 2 classes in the data, 
                                but the data contains only one class''')
        elif n_classes == 2:
            estimator = LogisticRegression(penalty = 'l1',
                                           class_weight = self.class_weight,
                                           C = 1/self.lambd,
                                           fit_intercept = self.fit_intercept,
                                           random_state = self.random_state,
                                           solver = self.solver,
                                           max_iter = self.max_iter)
            estimator.fit(X, Y)
            zero_indexes = np.where(np.isclose(estimator.coef_[0], 0))[0]
            _X = np.copy(X)
            _X[:,zero_indexes] = 0
            estimator = LogisticRegression(penalty = 'l1',
                                           class_weight = self.class_weight,
                                           C = 1/(self.lambd*self.phi),
                                           fit_intercept = self.fit_intercept,
                                           random_state = self.random_state,
                                           solver = self.solver,
                                           max_iter = self.max_iter)
            estimator.fit(_X, Y)
            self.intercept.append(estimator.intercept_[0]) 
            self.beta.append(estimator.coef_[0])
        else:    
            for j,k in enumerate(self.classes):
                _Y = np.where(Y == k, 1, 0)
                estimator = LogisticRegression(penalty = 'l1',
                                               class_weight = self.class_weight,
                                               C = 1/self.lambd,
                                               fit_intercept = self.fit_intercept,
                                               random_state = self.random_state,
                                               solver = self.solver,
                                               max_iter = self.max_iter)
                estimator.fit(X, _Y)
                zero_indexes = np.where(np.isclose(estimator.coef_[0], 0))[0]
                _X = np.copy(X)
                _X[:,zero_indexes] = 0
                estimator = LogisticRegression(penalty = 'l1',
                                               class_weight = self.class_weight,
                                               C = 1/(self.lambd*self.phi),
                                               fit_intercept = self.fit_intercept,
                                               random_state = self.random_state,
                                               solver = self.solver,
                                               max_iter = self.max_iter)
                estimator.fit(_X, _Y)
                self.intercept.append(estimator.intercept_[0]) 
                self.beta.append(estimator.coef_[0])
        self.intercept = np.array(self.intercept)
        self.beta = np.array(self.beta)            
        return self



class RelaxedLassoRegressor(BaseRegressor):
    def __init__(self,
                 fit_intercept: bool = True,
                 scale: bool = True,
                 random_state: int = None,
                 selection: str = 'cyclic',
                 max_iter: int = 1000):
        self.fit_intercept = fit_intercept
        self.parameters = {}
        self.lambd = None
        self.phi = None
        self.beta = None
        self.intercept = None
        self.estimator = None
        self.scale = scale
        self.scaler = None
        self.random_state = random_state
        self.max_iter = max_iter
        self.selection = selection

    
    def set_parameters(self,
                       parameters: Dict):
        self.phi = parameters['phi']
        self.lambd = parameters['lambda']
        self.parameters['phi'] = self.phi
        self.parameters['lambda'] = self.lambd
        if self.lambd <= 0:
            raise ValueError('Lambda <= 0')
        if self.phi<=0 or self.phi>1:
            raise ValueError('Phi out of (0,1]')
        
                
    def predict(self,
                X: np.ndarray) -> np.ndarray:
        if self.scale:
            X = self.scaler.transform(X)
        return self.estimator.predict(X)
                
                
    def fit(self,
            X: np.ndarray,
            Y: np.ndarray) -> 'RelaxedLassoRegressor':          
        if self.scale:
            self.scaler = StandardScaler()
            X = self.scaler.fit_transform(X)
        estimator = Lasso(alpha = self.lambd,
                          fit_intercept = self.fit_intercept,
                          random_state = self.random_state,
                          selection = self.selection,
                          max_iter = self.max_iter)
        estimator.fit(X, Y)
        zero_indexes = np.where(np.isclose(estimator.coef_, 0))[0]
        X[:,zero_indexes] = 0
        estimator = Lasso(alpha = self.phi*self.lambd,
                          fit_intercept = self.fit_intercept,
                          random_state = self.random_state,
                          selection = self.selection,
                          max_iter = self.max_iter)
        self.estimator = estimator.fit(X, Y)
        self.intercept = estimator.intercept_ 
        self.beta = estimator.coef_
        return self
        
        



import joblib
import numpy as np
import pyomo.environ as pyo

from pyomo.core.expr import numeric_expr
from pyomo.opt import SolverFactory
from sklearn.preprocessing import StandardScaler
from sklearn.linear_model import Ridge, LogisticRegression
from penlm.base_estimators import BaseClassifier, BaseRegressor
from abc import ABC, abstractmethod
from typing import Dict, Tuple



class BaseSACR(ABC):
    def set_parameters(self,
                       parameters: Dict):
        self.parameters['lambda'] = parameters['lambda']
        self.parameters['phi'] = parameters['phi']
        self.lambd = parameters['lambda']
        self.phi = parameters['phi']
        if self.lambd <= 0:
            raise ValueError('Lambda <= 0')
        if self.phi<=0 or self.phi>1:
            raise ValueError('Phi out of (0,1]')


    @abstractmethod
    def _init_ridge(self,
                    X: np.ndarray,
                    Y: np.ndarray):
        pass
            
                            
    def _joint_ridge(self,
                     X: np.ndarray,
                     Y: np.ndarray,
                     g: np.ndarray) -> Tuple[int,np.ndarray,np.ndarray]:
        N = X.shape[0]
        P = X.shape[1]
        model = pyo.ConcreteModel()
        model.P = pyo.RangeSet(0, P-1)
        model.beta = pyo.Var(model.P, 
                             domain = pyo.Reals)
        model.w = pyo.Var(model.P, 
                          domain = pyo.NonNegativeReals)
        if self.fit_intercept:
            model.intercept = pyo.Var(domain = pyo.Reals,
                                      initialize = 0)
        for p in range(P):
            model.beta[p]= 0
            model.w[p]= 0
                    
        def constraint_w_integral(model):
            return (1/P)*sum(model.w[p] for p in model.P) == 1
        model.constraint_w_integral = pyo.Constraint(rule = constraint_w_integral)
        
        model.obj = pyo.Objective(expr = self._obj_expr(model, X, Y, g),
                                  sense = pyo.minimize)
        opt = SolverFactory(self.solver,
                            options={'max_iter': self.max_iter}) 
        opt.solve(model)
        beta = []
        w = []
        for p in range(P):
            beta.append(pyo.value(model.beta[p]))
            w.append(pyo.value(model.w[p]))   
        beta = np.array(beta)
        w = np.array(w)
        if self.fit_intercept:
            intercept = pyo.value(model.intercept)
        else:
            intercept = 0
        return intercept, beta, w
        

    @abstractmethod
    def _obj_expr(self,
                  model: pyo.ConcreteModel,
                  X: np.ndarray,
                  Y: np.ndarray,
                  g: np.ndarray) -> numeric_expr:
        pass


    def _get_first_penalty_expr(self,
                                model: pyo.ConcreteModel,
                                P: int,
                                g: np.ndarray) -> numeric_expr:
        return sum((model.beta[p] - g[p]*model.w[p])**2 for p in range(P))
    
    
    def _get_second_penalty_expr(self,
                                 model: pyo.ConcreteModel,
                                 P: int,
                                 g: np.ndarray) -> numeric_expr:
        return P*P*sum((g[p]*model.w[p] - 2*g[p+1]*model.w[p+1] + g[p+2]*model.w[p+2])**2 for p in range(P-2))
        
        
            
class SACRClassifier(BaseSACR, BaseClassifier):
    def __init__(self,
                 solver: str = 'ipopt',
                 scikit_solver: str = 'saga',
                 fit_intercept: bool = True,
                 max_iter: int = 100,
                 scikit_max_iter: int = 1000,
                 scale: bool = True,
                 class_weight: str = 'balanced'):
        self.solver = solver
        self.scikit_solver = scikit_solver
        self.fit_intercept = fit_intercept
        self.max_iter = max_iter
        self.scikit_max_iter = scikit_max_iter
        self.scale = scale
        self.class_weight = class_weight
        self.parameters = {}
        self.beta = []
        self.w = []
        self.g = []
        self.intercept = []
        self.lambd = None
        self.phi = None
        self.scaler = None
        self.classes = None


    def fit(self,
            X: np.ndarray,
            Y: np.ndarray) -> 'SACRClassifier':
        if self.scale:
            self.scaler = StandardScaler()
            X = self.scaler.fit_transform(X)            
        self.classes = np.unique(Y)
        n_classes = len(self.classes)
        if n_classes < 2:
            raise ValueError('''This estimator needs samples of at least 2 classes in the data, 
                                but the data contains only one class''')
        elif n_classes == 2:
            g = self._init_ridge(X, Y)
            intercept, beta, w = self._joint_ridge(X, Y, g)
            self.intercept.append(intercept)
            self.beta.append(beta)
            self.w.append(w)
            self.g.append(g)  
        else:
            for j,k in enumerate(self.classes):
                _Y = np.where(Y == k, 1, 0)
                g = self._init_ridge(X, _Y)
                intercept, beta, w = self._joint_ridge(X, _Y, g)
                self.intercept.append(intercept)
                self.beta.append(beta)
                self.w.append(w)
                self.g.append(g)
        self.intercept = np.array(self.intercept)
        self.beta = np.array(self.beta)
        self.w = np.array(self.w)
        self.g = np.array(self.g)            
        return self
        
        
    def _init_ridge(self,
                    X: np.ndarray,
                    Y: np.ndarray) -> np.ndarray:
        estimator = LogisticRegression(penalty = 'l2',
                                       C = 1/self.lambd,
                                       class_weight = self.class_weight,
                                       fit_intercept = self.fit_intercept,
                                       solver = self.scikit_solver,
                                       max_iter = self.scikit_max_iter)
        estimator.fit(X, Y)
        return estimator.coef_[0]


    def _obj_expr(self,
                  model: pyo.ConcreteModel,
                  X: np.ndarray,
                  Y: np.ndarray,
                  g: np.ndarray) -> numeric_expr:
        N = X.shape[0]
        P = X.shape[1]
        if self.class_weight=='balanced': 
            unique_weights = N/(len(np.unique(Y))*np.bincount(Y))
            weights = np.zeros(N)
            for i in np.unique(Y):
                weights[np.where(Y==i)] = unique_weights[i]
        else:
            weights = np.ones(N)
        if self.fit_intercept:
            obj = -sum(weights[j]*(Y[j]*(sum(X[j,p]*model.beta[p] for p in range(P))+model.intercept) - pyo.log(1+pyo.exp(model.intercept+sum(X[j,p]*model.beta[p] for p in range(P)))))  for j in range(N))
        else:
            obj = -sum(weights[j]*(Y[j]*(sum(X[j,p]*model.beta[p] for p in range(P))) - pyo.log(1+pyo.exp(sum(X[j,p]*model.beta[p] for p in range(P)))))  for j in range(N))
        pen1 = self._get_first_penalty_expr(model, P, g)
        pen2 = self._get_second_penalty_expr(model, P, g)
        return obj + self.phi*self.lambd*pen1 + (1-self.phi)*self.lambd*pen2
                



class SACRRegressor(BaseSACR, BaseRegressor):
    def __init__(self,
                 solver: str = 'ipopt',
                 fit_intercept: bool = True,
                 max_iter: int = 100,
                 scikit_max_iter: int = 1000,
                 scale: bool = True):
        self.solver = solver
        self.fit_intercept = fit_intercept
        self.max_iter = max_iter
        self.scikit_max_iter = scikit_max_iter
        self.scale = scale
        self.parameters = {}
        self.beta = None
        self.w = None
        self.g = None
        self.intercept = None
        self.lambd = None
        self.phi = None
        self.scaler = None


    def fit(self,
            X: np.ndarray,
            Y: np.ndarray) -> 'SACRRegressor':
        self.classes = np.unique(Y)    
        if self.scale:
            self.scaler = StandardScaler()
            X = self.scaler.fit_transform(X)
        self.g = self._init_ridge(X, Y)
        self.intercept, self.beta, self.w = self._joint_ridge(X, Y, self.g)       
        return self
        
        
    def _init_ridge(self,
                    X: np.ndarray,
                    Y: np.ndarray) -> np.ndarray:
        estimator = Ridge(alpha = self.lambd,
                          fit_intercept = self.fit_intercept,
                          max_iter = self.scikit_max_iter)
        estimator.fit(X, Y)
        return estimator.coef_


    def _obj_expr(self,
                  model: pyo.ConcreteModel,
                  X: np.ndarray,
                  Y: np.ndarray,
                  g: np.ndarray) -> numeric_expr:
        N = X.shape[0]
        P = X.shape[1]
        if self.fit_intercept:
            obj = sum((Y[j]-sum(X[j,p]*model.beta[p] for p in range(P))-model.intercept)**2 for j in range(N))
        else:
            obj = sum((Y[j]-sum(X[j,p]*model.beta[p] for p in range(P)))**2 for j in range(N))
        pen1 = self._get_first_penalty_expr(model, P, g)
        pen2 = self._get_second_penalty_expr(model, P, g)
        return obj + self.phi*self.lambd*pen1 + (1-self.phi)*self.lambd*pen2






import numpy as np
import copy

from sklearn.preprocessing import StandardScaler      
from sklearn.linear_model import Lasso, LogisticRegression
from penlm.base_estimators import BaseClassifier, BaseRegressor
from typing import Dict
              
                                                                          
class AdaptiveLassoClassifier(BaseClassifier):
    def __init__(self,
                 fit_intercept: bool = True,
                 class_weight: str = 'balanced',
                 scale: bool = True,
                 random_state: int = None,
                 solver: str = 'saga',
                 max_iter: int = 1000):
        self.class_weight = class_weight
        self.solver = solver 
        self.fit_intercept = fit_intercept
        self.parameters = {}
        self.lambd = None
        self.gamma = None
        self.beta_init = None
        self.beta = []
        self.intercept = []
        self.scale = scale
        self.scaler = None
        self.random_state = random_state
        self.max_iter = max_iter
        self.classes = None

    
    def set_parameters(self,
                       parameters: Dict):
        self.gamma = parameters['gamma']
        self.lambd = parameters['lambda']
        self.beta_init = parameters['beta_init'][1]
        self.parameters['gamma'] = self.gamma
        self.parameters['lambda'] = self.lambd
        self.parameters['beta_init'] = parameters['beta_init'][0] 
        if self.lambd <= 0:
            raise ValueError('Lambda <= 0')
        if self.gamma <= 0:
            raise ValueError('Gamma <= 0')
                            
                
    def fit(self,
            X: np.ndarray,
            Y: np.ndarray) -> 'AdaptiveLassoClassifier':
        if self.scale:
            self.scaler = StandardScaler()
            X = self.scaler.fit_transform(X)            
        self.classes = np.unique(Y)
        n_classes = len(self.classes)
        if n_classes < 2:
            raise ValueError('''This estimator needs samples of at least 2 classes in the data, 
                                but the data contains only one class''')
        elif n_classes == 2:
            recipr_w = np.power(np.abs(self.beta_init[0]),
                                self.gamma)
            _X = X * np.tile(recipr_w, 
                             (X.shape[0], 1))
            estimator = LogisticRegression(penalty = 'l1',
                                           class_weight = self.class_weight,
                                           C = 1/self.lambd,
                                           fit_intercept = self.fit_intercept,
                                           random_state = self.random_state,
                                           solver = self.solver,
                                           max_iter = self.max_iter)
            estimator.fit(_X, Y)
            self.intercept.append(estimator.intercept_[0])
            self.beta.append(estimator.coef_[0]*recipr_w)
        else:    
            for j,k in enumerate(self.classes):
                recipr_w = np.power(np.abs(self.beta_init[j]),
                                    self.gamma)
                _X = X * np.tile(recipr_w, 
                                 (X.shape[0], 1))
                _Y = np.where(Y == k, 1, 0)     
                estimator = LogisticRegression(penalty = 'l1',
                                               class_weight = self.class_weight,
                                               C = 1/self.lambd,
                                               fit_intercept = self.fit_intercept,
                                               random_state = self.random_state,
                                               solver = self.solver,
                                               max_iter = self.max_iter)
                estimator.fit(_X, _Y)
                self.intercept.append(estimator.intercept_[0]) 
                self.beta.append(estimator.coef_[0]*recipr_w)
        self.intercept = np.array(self.intercept)
        self.beta = np.array(self.beta)
        return self


class AdaptiveLassoRegressor(BaseRegressor):
    def __init__(self,
                 fit_intercept: bool = True,
                 scale: bool = True,
                 random_state: int = None,
                 selection: str = 'cyclic',
                 max_iter: int = 1000):
        self.fit_intercept = fit_intercept
        self.parameters = {}
        self.lambd = None
        self.gamma = None
        self.beta_init = None
        self.beta = None
        self.intercept = None
        self.scale = scale
        self.scaler = None
        self.random_state = random_state
        self.selection = selection
        self.max_iter = max_iter

    
    def set_parameters(self,
                       parameters: Dict):
        self.gamma = parameters['gamma']
        self.lambd = parameters['lambda']
        self.beta_init = parameters['beta_init'][1]
        self.parameters['gamma'] = self.gamma
        self.parameters['lambda'] = self.lambd
        self.parameters['beta_init'] = parameters['beta_init'][0] 
        if self.lambd <= 0:
            raise ValueError('Lambda <= 0')
        if self.gamma <= 0:
            raise ValueError('Gamma <= 0')
                            
                
    def fit(self,
            X: np.ndarray,
            Y: np.ndarray) -> 'AdaptiveLassoRegressor':         
        if self.scale:
            self.scaler = StandardScaler()
            X = self.scaler.fit_transform(X)
        recipr_w = np.power(np.abs(self.beta_init),
                            self.gamma)
        for i in range(X.shape[0]):
            X[i,:] = X[i,:]*recipr_w
        estimator = Lasso(alpha = self.lambd,
                          fit_intercept = self.fit_intercept,
                          random_state = self.random_state,
                          max_iter = self.max_iter,
                          selection = self.selection)
        estimator.fit(X, Y)
        self.intercept = estimator.intercept_ 
        self.beta = estimator.coef_*recipr_w
        return self
        
        



from penlm.grid_search import GridSearchCV
from penlm.base_estimators import BaseClassifier, BaseRegressor
from penlm.adaptive_lasso import AdaptiveLassoClassifier, AdaptiveLassoRegressor
from penlm.bar_estimator import BARClassifier, BARRegressor
from penlm.non_negative_garrote import NNGarroteClassifier, NNGarroteRegressor
from penlm.relaxed_lasso import RelaxedLassoClassifier, RelaxedLassoRegressor
from penlm.smooth_linear_model import SmoothLinearClassifier, SmoothLinearRegressor
from penlm.smoothly_adaptively_centered_ridge import SACRClassifier, SACRRegressor




from setuptools import setup, find_packages
from os import path

# read the contents of your README file
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
  name = 'penlm',
  packages = ['penlm'],
  version = 'v1.0.12', 
  license='MIT',
  description = 'Penalized Linear Models for Classification and Regression',
  long_description=long_description,
  long_description_content_type='text/markdown',
  author = 'Edoardo Belli',
  author_email = 'iedobelli@gmail.com',
  url = 'https://github.com/bellibot/penlm',
  download_url = 'https://github.com/bellibot/penlm/archive/refs/tags/v1.0.12.tar.gz',
  keywords = ['Classification', 'Regression', 'Linear', 'Penalty'],
  python_requires='>=3.5',
  install_requires=[
          'numpy',
          'joblib',
          'Pyomo',
          'scikit-learn',
      ],
  classifiers=[
    'Development Status :: 5 - Production/Stable',   # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable"
  ],
)



import numpy as np

from typing import Dict, Callable
from abc import ABC, abstractmethod
from sklearn.metrics import accuracy_score, r2_score
                                  

class BaseClassifier(ABC):
    estimator_type = 'classifier'
    
    
    def predict(self,
                X: np.ndarray) -> np.ndarray:
        probs = self.predict_proba(X)
        return np.argmax(probs, axis=1)
        
        
    def predict_proba(self,
                      X: np.ndarray) -> np.ndarray:           
        if self.scale:
            X = self.scaler.transform(X)
        intercepts = np.tile(self.intercept, 
                             (X.shape[0], 1))                   
        log_odds = np.matmul(X, self.beta.transpose()) + intercepts
        probs_one = (1 / (1 + np.exp(-log_odds)))        
        if len(self.classes) == 2:
            probs = np.vstack([1 - probs_one[:,0], probs_one[:,0]]).T
        else:
            probs = probs_one/probs_one.sum(axis=1).reshape((probs_one.shape[0], -1))
        return probs
                
                
    def score(self,
              X: np.ndarray,
              Y: np.ndarray,
              scoring: Callable = None) -> np.ndarray:
        pred_Y = self.predict(X) 
        if scoring == None:
            score = accuracy_score(Y, pred_Y) 
        elif callable(scoring):
            score = scoring(Y, pred_Y)
        else:
            raise ValueError(f'Illegal scoring {scoring}')          
        return score
        
        
    @abstractmethod
    def set_parameters(self,
                       parameters: Dict):
        pass
          
                
    @abstractmethod            
    def fit(self,
            X: np.ndarray,
            Y: np.ndarray) -> 'BaseClassifier':          
        pass    
        
        
        
class BaseRegressor(ABC):
    estimator_type = 'regressor'
    
    
    def predict(self,
                X: np.ndarray) -> np.ndarray:
        if self.scale:
            X = self.scaler.transform(X)
        pred_Y = []
        for x in X:
            pred = np.dot(self.beta,x) + self.intercept
            pred_Y.append(pred)
        return np.array(pred_Y)
        
            
    def score(self,
              X: np.ndarray,
              Y: np.ndarray,
              scoring: Callable = None) -> np.ndarray:
        pred_Y = self.predict(X) 
        if scoring == None:
            score = r2_score(Y, pred_Y) 
        elif callable(scoring):
            score = scoring(Y, pred_Y)
        else:
            raise ValueError(f'Illegal scoring {scoring}')          
        return score
        
        
    @abstractmethod
    def set_parameters(self,
                       parameters: Dict):
        pass
          
                
    @abstractmethod            
    def fit(self,
            X: np.ndarray,
            Y: np.ndarray) -> 'BaseRegressor':          
        pass    




